"""
This file is the entrypoint for the cli_handler package.
"""
import asyncio
from datetime import datetime, timedelta, timezone
import json
from pprint import pprint
import typer
from extractor.settings import ValidModelsData, load_api_settings
from logger import logger_conf
import utils.constants as constants
from utils.current_run import CurrentRunConstants
from utils.last_succesful_run import Filters, LastRunFileManager, LastSuccessfulRun


# The application object that will run the extractor
app = typer.Typer()


def print_help(models_data: ValidModelsData):
    print("\n\nRun: python main.py [--model-name MODEL_NAME] [--date-from START_DATE --date-to END_DATE] [--debug]")
    print("Where:")
    print("\t --model-name should be one of:")
    for model in models_data.valid_models:
        print(f"\t\t + {model}")
        model_requirements = models_data.api_settings[model].get("requirements", None)
        if model_requirements is not None:
            mandatory = model_requirements.get("mandatory", None)
            if mandatory is not None:
                print("\t\t\t Mandatory arguments: ")
                for argument in mandatory:
                    print(f"\t\t\t\t --{constants.FILTERS_MAPPING_USER[argument]}")

    print("\n\n\t --date-from is a date in format YYYY-MM-DD")
    print("\t --date-to is a date in format YYYY-MM-DD")

def extract_model(
    model_name: str,
    output_dir: str,
) -> None:
    """
    Extract data from the API.

    Args:
        model_name (str): Name of the model to extract.        
        output_dir (str): Output directory for the extracted data.

    Returns:
        None
    """
    logger = logger_conf.get_logger("MAIN")    
    #read the configuration for the model "model_name" from the yaml files
    
    model_data = load_api_settings(model_name, output_dir)

    if not model_data.is_valid:
        if not model_data.is_valid_model_name:    
            print(f"\n'{model_name}' is not a valid model name.")
        if not model_data.are_constants_valid:
            print("\nPlease set up the next constants in the 'user_settings.yaml' file")
            for c in model_data.unset_constants:
                print(f"\t - {c}")
        print_help(model_data)        
        return
    
    current_run = CurrentRunConstants()
    

    filters_params = {}
    # If delta option is setup
    if current_run.get_value("delta"):
        if model_name in constants.DELTA_MODELS:
            #read last_run.json        
            last_run = LastRunFileManager()
            last_run_status = last_run.get_status(model_name)
            # if there are no entries for this model_name, exit with a warning
            if last_run_status is None:
                print(f"\n\t There is no previous succesful run for model {model_name}.")
                return         
            # if there is entry, set update_date_from with readed date
            # and update_date_to with today's date.
            else:
                # create a datetime object and add one second to the date
                update_date_from = last_run_status.get("latest_update",None)
                update_date_from = datetime.strptime(update_date_from, "%Y-%m-%dT%H:%M:%SZ")
                # Add one second because the server runs the query as 'greater than or equal'. So we would be retrieving data that didn't change
                update_date_from = update_date_from + timedelta(seconds=1)
                filters_params["dateFrom"] = update_date_from.strftime("%Y-%m-%dT%H:%M:%SZ")
                # Using UTC time zone, because they work with GMT (There is no time difference between Coordinated Universal Time and Greenwich Mean Time)
                filters_params["dateTo"] = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
    else:
        try:
            date_from = current_run.get_value("date_from")
            date_to = current_run.get_value("date_to")            
            if date_from is not None:
                date_from = datetime.strptime(date_from, "%Y-%m-%d")                
                filters_params["dateFrom"] = date_from.strftime("%Y-%m-%dT%H:%M:%SZ")
            if date_to is not None:
                date_to = datetime.strptime(date_to, "%Y-%m-%d")
                date_to = date_to.replace(hour=23, minute=59, second=59)
                filters_params["dateTo"] = date_to.strftime("%Y-%m-%dT%H:%M:%SZ")
        except ValueError as e:
            print("Error when parsing dates: ")
            print(f"\t {e}")
            return



    valid_filters, mandatory_filters = model_data.validate_filters(filters_params)
    if not valid_filters:
        print(f"Model '{model_name}' requires mandatory params")
        for filter in mandatory_filters:
            print(f"\t --{constants.FILTERS_MAPPING_USER[filter]}")
        return
    
    logger.info(f"Begin download model: '{model_name}'")

    # Start to download the data!
    print(f"Running {model_name} extraction.. ")
    last_run = LastSuccessfulRun(model_name)
    # instanciate a template_method with the selected model wether is correctly set or not
    # to assure creation of output files creation
    template_method = model_data.get_method_instance()
    template_method.set_last_run(last_run)
    asyncio.run(template_method.download_data())
    last_run = template_method.get_last_run()
    manager = LastRunFileManager()
    if last_run.latest_update != None:
        manager.save_status(model_name, last_run.get_run_status())
    print(f"Extraction of {model_name} finished")
    logger.info(f"Model download finished")


@app.command()
def extract(
    model_name: str = typer.Option(
        "", help="Assign the name of the api that you want to download."
    ),
    date_from: str = typer.Option(
        None, help="Data extraction start create date format: YYYY-MM-DD."
    ),
    date_to: str = typer.Option(
        None, help="Data extraction end create date format: YYYY-MM-DD."
    ),
    delta: bool = typer.Option(
        False, help="Option to download delta load for selected model."
    ),
    allp2p: bool = typer.Option(
        False, help="Option to download all P2P models"
    ),
    allsourcing: bool = typer.Option(
        False, help="Option to download all sourcing models"
    ),
    debug: bool = typer.Option(
        False, help="Set this option to have more deep logs."
    ),
    output: str = typer.Option("", help="Folder to save extracted data.")
):
    """Extract data from the API."""

    
    current_run = CurrentRunConstants()
    current_run.set_value("model_name", model_name)
    current_run.set_value("date_from", date_from)
    current_run.set_value("date_to", date_to)
    current_run.set_value("delta", delta)
    current_run.set_value("allp2p", allp2p)
    current_run.set_value("allsourcing", allsourcing)
    current_run.set_value("debug", debug)

    if output == "":
        output_dir = f'./output/{datetime.now().strftime("%Y_%m_%d_%H_%M_%S")}'
    else:
        output_dir = output

    logger_conf.setup_logger(output_dir, debug)
    if model_name:
        extract_model(model_name, output_dir)
    elif allp2p:
        for model in constants.P2P_MODELS:
            extract_model(model, output_dir)
    elif allsourcing:
        for model in constants.SOURCING_MODELS:
            extract_model(model, output_dir)
    else:
        print("Please choose one model with: --model-name option. Or one of --allp2p or --allsourcing.")