"""Settings for extractor."""
from datetime import datetime
import os
import utils.constants as constants
from utils.current_run import CurrentRunConstants
from utils.file_handling import file_exists
from utils.yaml_composing import load_composed_yaml

class ValidModelsData():
    def __init__(self, settings, model_name, output_dir):        
        self.model_name = model_name
        self.output_dir = output_dir
        self.constanst = settings.get("constants", None)
        self.api_settings = settings.get("api", None)
        self.valid_models = self.get_valid_models_list()
        self.is_valid_model_name = self.validate_model_name()
        self.are_constants_valid, self.unset_constants = self.validate_constants()
        self.is_valid = self.is_valid_model_name and self.are_constants_valid
        self.current_run = CurrentRunConstants()
        self.current_run.set_value("remove_old", self.constanst.get("remove_old", False))    
        self.current_run.set_value("filenames_as_timestamp", self.constanst.get("filenames_as_timestamp", False))
        if self.is_valid:
            self.current_model_settings = self.setup_model_settings(model_name)
        

    def create_output_files(self, model_settings):
        #first clean output files
        output_handler = model_settings.get("output_handler", None)
        tables = model_settings.get("tables",None)
        model = model_settings.get("model", None)
        models = model_settings.get("models", None)        
        if model is not None:
            self.create_file(tables, model, output_handler)
        if models is not None:
            for model in models:
                self.create_file(tables, model, output_handler)
    
    def create_file(self, tables, model, output_handler):
        output_handler = output_handler(
                tables=tables,
                model=model,
                output_dir=self.output_dir,
        )
        extension = output_handler.get_extension()
        for table_name, headers in tables.items():
            if not self.current_run.get_value(f"{table_name}_cleaned"):
                self.current_run.set_value(f"{table_name}_cleaned", True)
                file_name = output_handler.get_filename(table_name)
                full_file_name = f"{file_name}{extension}"
                if file_exists(full_file_name):
                    if self.current_run.get_value("remove_old"):
                        os.remove(full_file_name)
                    else:
                        timestamp = datetime.now().strftime("%Y%m%d%H%M%S%f")[:-3]
                        full_new_name = f"{file_name}_previous_to_{timestamp}{extension}"
                        os.rename(full_file_name, full_new_name)
                    
                    #after clean it, create it!
                    output_handler.create_file(full_file_name,headers)
                else:
                    #if doesn't exists create it
                    output_handler.create_file(full_file_name,headers)


    def get_valid_models_list(self):
        valid_models = []
        for model_key, model_config in self.api_settings.items():
            model_requirements = model_config.get("requirements", None)
            if model_requirements is not None:
                if model_requirements.get("independent", False):
                    valid_models.append(model_key)

        return valid_models
    
    def setup_model_settings(self, model_name):
        temp_model_settings = self.api_settings[model_name]
        model_settings = self.map_settings(temp_model_settings, model_name, self.output_dir)
        self.create_output_files(model_settings)
        list_extras = model_settings.get("extras", None)
        if list_extras is not None:
            model_settings["extras"] = {}
            for extra in list_extras:
                model_settings["extras"][extra] = self.setup_model_settings(extra)

        return model_settings


    def map_settings(self, settings: dict, model_name, output_dir):
        if isinstance(settings.get("model", None), str):
            # Enter only if this model hasn't been set before
            settings["model"] = constants.MODELS_MAPPING[settings["model"]]
        elif settings.get("models", None):
            models_list = settings.get("models")
            settings["models"] = {}
            for model in models_list:
                if isinstance(model, str):
                    settings["models"][model] = constants.MODELS_MAPPING[model]
        list_tables = settings.get("tables", [])
        settings["tables"] = {}
        for table in list_tables:
            settings["tables"][table] = constants.TABLES_MAPPING[table]
        if isinstance(settings.get("method", None), str):
            # Enter only if this method hasn't been set before
            settings["method"] = constants.METHODS_MAPPING[settings["method"]]
        if isinstance(settings.get("output_handler", None), str):
            settings["output_handler"] = constants.OUTPUT_HANDLERS_MAPPING[settings["output_handler"]]
        if isinstance(settings.get("view_name", None), str):
            settings["view_name"] = constants.PARSER_MAPPING[settings["view_name"]]
        settings["model_name"] = model_name
        settings["output_dir"] = output_dir  

        return settings
    

    def validate_model_name(self) -> bool:
        is_valid = True
        # Check if model_name is in list of valid models
        if not self.model_name in self.valid_models:
            is_valid = False 
        return is_valid
    
    def validate_filters(self, filters):
        # Check if model has mandatory filters and they were passed on
        valid_filters = True
        mandatory_filters = self.current_model_settings.get("requirements",{}).get("mandatory",[])
        for filter in mandatory_filters:
            if filter not in filters:
                valid_filters = False

        if valid_filters:
            self.set_filters(filters)

        return valid_filters, mandatory_filters

    def set_filters(self, filters_params):
        def setup_filters(temp_settings):
            possible_urls = ["base", "count", "list", "download","delta_list"]
            possible_filter_fields = ["filters"]
            for url in possible_urls:
                if temp_settings.get("params", {}).get(url, None):
                    for filter_field in possible_filter_fields:
                        temp_settings["params"][url][filter_field] = filters_params

            # check is current temp settings have extras
            for extra_model_key, extra_model_data in temp_settings.get("extras",{}).items():
                setup_filters(extra_model_data)


        setup_filters(self.current_model_settings)
        

    def get_method_instance(self):
        return self.current_model_settings["method"](self.current_model_settings)
    

    def validate_constants(self):
        """ By default constanst are defined with a value of '****'. This process is to check if the final user updated those values."""
        is_valid = True
        unset_constants = []
        for constant, value in self.constanst.items():
            if value == "****":
                is_valid = False
                unset_constants.append(constant)

        return is_valid, unset_constants

def load_api_settings(model_name: str, output_dir) -> ValidModelsData:
    """Load api settings from user_settings.yaml."""    

    original_settings = load_composed_yaml("user_settings.yaml")  

    #valid_models = get_valid_models_list(original_settings)
    #
    #if model_name not in valid_models:
    #    return ValidModelsData(f"Model Name: {model_name} is not a valid model", valid_models, original_settings)

    return ValidModelsData(original_settings, model_name, output_dir)


    

