"""Operational Reporting for Sourcing models."""
from datetime import datetime

from pydantic import BaseModel, Field, validator
from typing import Optional


class SourcingProject(BaseModel):
    """SourcingProject model."""

    InternalId: Optional[str]  # ok
    Title: Optional[str]  # ok
    Status: Optional[str]  # ok
    BaselineSpendCurrency: Optional[str] = Field(None, alias="BaselineSpend")  # ok
    BaselineSpendAmount: Optional[str] = Field(None, alias="BaselineSpend")  # ok
    OwnerName: Optional[str] = Field(None, alias="Owner")  # ok
    RegionDescription: Optional[str] = Field(None, alias="Region")  # ok
    ClientDescription: Optional[str] = Field(None, alias="Client")  # no
    CommodityName: Optional[str] = Field(
        None, alias="Commodity"
    )  # dentro de commodities
    TimeCreated: Optional[str]  # esta como createDate
    ParentDocumentInternalId: Optional[str] = Field(
        None, alias="ParentDocument"
    )  # dentro de parentIdsTitlesMap
    lastModified: Optional[str]

    @validator("lastModified", pre=True, always=True)
    def get_modified_date(cls, value):
        """Convert ISO 8601 to datetime"""
        if not value:
            return ""
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z").strftime(
            "%Y-%m-%d %H:%M:%S"
        )

    @validator("BaselineSpendCurrency", pre=True, always=True)
    def get_currency(cls, value):
        """Get currency from BaselineSpend field."""
        if not value:
            return ""
        return value.get("Currency", {}).get("UniqueName", None)

    @validator("BaselineSpendAmount", pre=True, always=True)
    def get_amount(cls, value):
        """Get amount from BaselineSpend field."""
        if not value:
            return ""
        return value.get("Amount", None)

    @validator("OwnerName", pre=True, always=True)
    def get_owner_name(cls, value):
        """Get owner name from Owner field."""
        if not value:
            return ""
        return value.get("Name", None)

    @validator("RegionDescription", pre=True, always=True)
    def get_region_description(cls, value):
        """ "Get region description from Region field."""
        if not value:
            return ""
        return value[0].get("Description")

    @validator("ClientDescription", pre=True, always=True)
    def get_client_description(cls, value):
        """Get client description from Client field."""
        if not value:
            return ""
        return value[0].get("Description", None)

    @validator("CommodityName", pre=True, always=True)
    def get_commodity_name(cls, value):
        """Get commodity name from Commodity field."""
        if not value:
            return ""
        return value[0].get("Name", None)

    @validator("ParentDocumentInternalId", pre=True, always=True)
    def get_parent_document_id(cls, value):
        """Get parent document id from ParentDocument field."""
        if not value:
            return ""
        return value.get("InternalId", None)

    @validator("TimeCreated", pre=True, always=True)
    def get_time_created(cls, value):
        """Get time created from TimeCreated field and formats to ISO 8601."""
        if not value:
            return ""
        return datetime.fromisoformat(value.replace("Z", "+00:00")).strftime(
            "%Y-%m-%d %H:%M:%S"
        )


class SourcingProjectDelta(BaseModel):
    """SourcingProject model."""

    InternalId: Optional[str] = Field(None, alias="internalId")  # ok
    Title: Optional[str] = Field(None, alias="title")  # ok
    Status: Optional[str] = Field(None, alias="status")  # ok
    BaselineSpendCurrency: Optional[str] = Field(None, alias="baselineSpend")  # ok
    BaselineSpendAmount: Optional[str] = Field(None, alias="baselineSpend")  # ok
    OwnerName: Optional[str] = Field(None, alias="owner")  # ok
    RegionDescription: Optional[str] = Field(None, alias="regions")  # ok
    ClientDescription: Optional[str] = Field(None, alias="Client")  # no
    CommodityName: Optional[str] = Field(
        None, alias="commodities"
    )  # dentro de commodities
    TimeCreated: Optional[str] = Field(None, alias="createDate")  # esta como createDate
    ParentDocumentInternalId: Optional[str] = Field(
        None, alias="parentIdsTitlesMap"
    )  # dentro de parentIdsTitlesMap
    lastModified: Optional[str]

    @validator("lastModified", pre=True, always=True)
    def get_modified_date(cls, value):
        """Convert ISO 8601 to datetime"""
        if not value:
            return ""
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z").strftime(
            "%Y-%m-%d %H:%M:%S"
        )

    @validator("BaselineSpendCurrency", pre=True, always=True)
    def get_currency(cls, value):
        """Get currency from BaselineSpend field."""
        if not value:
            return ""
        return value.get("currency", "")

    @validator("BaselineSpendAmount", pre=True, always=True)
    def get_amount(cls, value):
        """Get amount from BaselineSpend field."""
        if not value:
            return ""
        return value.get("amount", None)

    @validator("OwnerName", pre=True, always=True)
    def get_owner_name(cls, value):
        """Get owner name from Owner field."""
        if not value:
            return ""
        return value.get("name", None)

    @validator("RegionDescription", pre=True, always=True)
    def get_region_description(cls, value):
        """ "Get region description from Region field."""
        if not value:
            return ""
        return value[0].get("name")

    @validator("ClientDescription", pre=True, always=True)
    def get_client_description(cls, value):
        """Get client description from Client field."""
        if not value:
            return ""
        return value[0].get("Description", None)

    @validator("CommodityName", pre=True, always=True)
    def get_commodity_name(cls, value):
        """Get commodity name from Commodity field."""
        if not value:
            return ""
        return value[0].get("name", None)

    @validator("ParentDocumentInternalId", pre=True, always=True)
    def get_parent_document_id(cls, value):
        """Get parent document id from ParentDocument field."""
        if not value:
            return ""
        return value[0].get("key", None)

    @validator("TimeCreated", pre=True, always=True)
    def get_time_created(cls, value):
        """Get time created from TimeCreated field and formats to ISO 8601."""
        if not value:
            return ""
        return datetime.fromisoformat(value.replace("Z", "+00:00")).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
