"""Sourcing Project Managment Models."""
from datetime import datetime

import pytz
from pydantic import BaseModel, Field, validator
from typing import Optional


class SourcingProjectEvent(BaseModel):
    """Sourcing Project Events Model."""

    internalId: Optional[str]
    description: Optional[str]
    projectId: Optional[str] = Field(None, alias="detail_id")
    type: Optional[str]
    docType: Optional[str]
    iconType: Optional[str]
    title: Optional[str]
    status: Optional[str]
    createDate: Optional[str]
    owner: Optional[str]
    parentId: Optional[str]
    parentProjectId: Optional[str]
    parentIdsTitlesMapKey1: Optional[str] = Field(None, alias="parentIdsTitlesMap")
    parentIdsTitlesMapKey2: Optional[str] = Field(None, alias="parentIdsTitlesMap")
    lastModified: Optional[str]

    @validator("lastModified", pre=True, always=True)
    def get_modified_date(cls, value):
        """Convert ISO 8601 to datetime"""
        if not value:
            return ""
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z").strftime(
            "%Y-%m-%d %H:%M:%S"
        )

    @validator("createDate", pre=True, always=True)
    def get_creation_date(cls, value):
        """Convert ISO 8601 to datetime"""
        if not value:
            return ""
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z").strftime(
            "%Y-%m-%d %H:%M:%S.%f"
        )[:-3]

    @validator("parentIdsTitlesMapKey1", pre=True, always=True)
    def get_parent_ids_titles_map_key1(cls, value):
        if not value:
            return ""
        try:
            return value[0].get("key", None)
        except IndexError:
            return ""

    @validator("parentIdsTitlesMapKey2", pre=True, always=True)
    def get_parent_ids_titles_map_key2(cls, value):
        if not value:
            return ""
        try:
            return value[1].get("key", None)
        except IndexError:
            return ""

    @validator("owner", pre=True, always=True)
    def get_owner(cls, value):
        if not value:
            return ""
        return value.get("name", None)


class SourcingProjectDetail(BaseModel):
    """Sourcing Project Detail Model."""

    internalId: Optional[str]
    projectId: Optional[str] = Field(None, alias="detail_id")
    templateProjectTitle: Optional[str]
    plannedStartDate: Optional[str]
    plannedEndDate: Optional[str]
    endDate: Optional[str]
    beginDate: Optional[str]
    lastModified: Optional[str]

    @validator("lastModified", pre=True, always=True)
    def get_modified_date(cls, value):
        """Convert ISO 8601 to datetime"""
        if not value:
            return ""
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z").strftime(
            "%Y-%m-%d %H:%M:%S"
        )

    @validator("plannedStartDate", pre=True, always=True)
    def get_planned_start_date(cls, value):
        """Convert ISO 8601 to datetime"""
        if not value:
            return ""
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z").strftime(
            "%Y-%m-%d %H:%M:%S.%f"
        )[:-3]

    @validator("plannedEndDate", pre=True, always=True)
    def get_planned_end_date(cls, value):
        """Convert ISO 8601 to datetime"""
        if not value:
            return ""
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z").strftime(
            "%Y-%m-%d %H:%M:%S.%f"
        )[:-3]

    @validator("endDate", pre=True, always=True)
    def get_end_date(cls, value):
        """Convert ISO 8601 to datetime"""
        if not value:
            return ""
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z").strftime(
            "%Y-%m-%d %H:%M:%S.%f"
        )[:-3]

    @validator("beginDate", pre=True, always=True)
    def get_begin_date(cls, value):
        """Convert ISO 8601 to datetime"""
        if not value:
            return ""
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f%z").strftime(
            "%Y-%m-%d %H:%M:%S.%f"
        )[:-3]


class SourcingProjectsHistory(BaseModel):
    """Sourcing Projects History."""

    id: Optional[str]
    ProjectId: Optional[str] = Field(None, alias="detail_id")
    title: Optional[str]
    timestamp: Optional[str] = Field(None, alias="timestamp")
    realUserName: Optional[str] = Field(None, alias="realUser")
    description: Optional[str] = Field(None, alias="description")
    type: Optional[str]
    details: Optional[str]

    class Config:
        allow_population_by_field_name = True


    @validator("realUserName", pre=True, always=True)
    def get_real_username(cls, value):
        """Get real user name."""
        if not value:
            return ""
        return value.get("name", None)

    @validator("timestamp", pre=True, always=True)
    def get_timestamp(cls, value):
        """Convert timestamp milliseconds to datetime"""
        if not value:
            return ""
        gmt = pytz.timezone("GMT")
        return datetime.fromtimestamp(value / 1000, gmt).strftime(
            "%Y-%m-%d %H:%M:%S.%f"
        )[:-3]

    @validator("description", pre=True, always=True)
    def get_description(cls, value):
        """Transform a list to a string"""
        if not value:
            return ""
        return ", ".join(value)
