"""Module to handle the creation of csv files."""
import csv
from datetime import datetime
import os.path
from csv import DictWriter
from typing import List

from pydantic import BaseModel

from utils.current_run import CurrentRunConstants



class CSVBase:
    def __init__(self, tables: dict, model: BaseModel, output_dir: str):
        """
        Initializes the class.

        Args:
            tables (dict): Dictionary with the name of the file and the headers.
            model (BaseModel): Pydantic model to validate the payload.
            output_dir (str): String where the extracted files are to be saved.
        """
        self.tables = tables
        self.model = model
        self.output_dir = output_dir

    
    def get_extension(self):
        return ".csv"
    
    def get_filename(self, table_name):
        current_run = CurrentRunConstants()
        if current_run.get_value("filenames_as_timestamp"):
            #mapping the file name to its datetime
            final_name = current_run.get_value(table_name)            
            if final_name is None:
                final_name = datetime.now().strftime("%Y%m%d%H%M%S%f")[:-3]
                # Check if folder exists, if not... create it
                new_path = os.path.join(self.output_dir, table_name)
                if not os.path.exists(f"{new_path}"):
                    os.makedirs(f"{new_path}")
                final_name = os.path.join(table_name, final_name)
                current_run.set_value(table_name, final_name)
        else:
            final_name = table_name
        final_name = os.path.join(self.output_dir, final_name)
        return final_name



    def create_file(self, file_name, headers):
        #current_run = CurrentRunConstants()
        #if not os.path.exists(self.output_dir):
        #    os.makedirs(self.output_dir)
        #if current_run.get_value("delta") and current_run.get_value("filenames_as_timestamp"):
        #        #mapping the file name to its datetime
        #        final_name = current_run.get_value(file_name)            
        #        if final_name is None:
        #            final_name = datetime.now().strftime("%Y%m%d%H%M%S%f")[:-3]
        #            # Check if folder exists, if not... create it
        #            new_path = os.path.join(self.output_dir, file_name)
        #            if not os.path.exists(f"{new_path}"):
        #                os.makedirs(f"{new_path}")
        #            final_name = os.path.join(file_name, final_name)
        #            current_run.set_value(file_name, final_name)
        #else:
        #    final_name = file_name
        #final_name = os.path.join(self.output_dir, final_name)
        with open(file_name, "+a", encoding='UTF-8', newline="") as file:
                writer = DictWriter(file, headers, delimiter="\t", quoting=csv.QUOTE_ALL )
                writer.writeheader()                
                print(f"{file_name} file created")

    def add_rows(self, payload: List[BaseModel], detail_id: str | None = None) -> None:
        """
        Creates a csv file with the headers provided.

        Args:
            payload (List[BaseModel]): List of pydantic models.
            detail_id (str | None): Id of the detail record.

        Returns:
            None
        """
        current_run = CurrentRunConstants()
        extension = self.get_extension()
        
        for table_name, headers in self.tables.items():
            #if current_run.get_value("delta"):
            #    #mapping the file name to its datetime
            #    final_name = current_run.get_value(file_name)
            #else:
            #    final_name = file_name
            final_name = self.get_filename(table_name)
            payload = self.clean_payload(payload, headers, detail_id)
            #final_name = os.path.join(self.output_dir, file_name)
            with open(f"{final_name}{extension}", "+a", encoding='UTF-8', newline="") as file:
                writer = DictWriter(file, headers, delimiter="\t", quoting=csv.QUOTE_ALL )
                for record in payload:
                    writer.writerow(record)

    def clean_payload(
        self,
        payload: List[BaseModel] | BaseModel,
        headers: List[str],
        detail_id: str | None = None,
    ) -> List[dict]:
        """
        Cleans the payload to match the headers.

        Args:
            payload (List[BaseModel] | BaseModel): List of pydantic models.
            headers (List[str]): List of headers.
            detail_id (str | None): Id of the detail record.
        """
        new_payload = []
        if isinstance(payload, BaseModel):
            payload = [payload]
        for record in payload:
            if not isinstance(record, dict) and isinstance(record, BaseModel):
                record = record.dict()
            for key in list(record.keys()):
                if key not in headers:
                    del record[key]
            new_payload.append(record)
        return new_payload
