from base64 import b64encode
from datetime import datetime, timedelta, timezone
from typing import Generator

from httpx import Auth, Response, AsyncClient, Client, Request

import logging

import utils.constants as constants


class AuthMeta(type):
    """Singleton metaclass for Auth class."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Check if the instance already exists."""
        is_new_instance = (
            cls not in cls._instances
            or cls._instances[cls].authorization != kwargs["authorization"]
        )
        if is_new_instance:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class AribaAuth(Auth, metaclass=AuthMeta):
    """
    Singleton Ariba auth class.

    This class is used to authenticate the requests to the Ariba API, using
    the OAuth2 protocol.

    The class is a singleton, so it will only be instantiated once, and
    the tokens will be refreshed when needed.
    """

    requires_response_body = True
    access_token: str = None
    refresh_token: str = None
    expiration_time: datetime = None

    def __init__(
        self,
        authorization: dict | None = None,
    ):
        self.authorization = authorization

    async def async_auth_flow_old(self, request):
        """
        Async auth flow.

        This method is called by the httpx client before sending the request.
        It will check if the access token is valid, and if not, it will
        refresh it.
        """
        if not self.access_token:
            tokens = self.sync_get_token()
            self.access_token = tokens["access_token"]
            self.refresh_token = tokens["refresh_token"]
            self.expiration_time = datetime.now(timezone.utc) + timedelta(seconds=tokens["expires_in"])
            # This is the response when authenticating
            #{
            #   "access_token": "42ba0187-852c-4be8-9df7-5b070ae52b9f",
            #   "refresh_token": "e12e49bf-a674-4e7f-a3a4-6b461bb9a6e7",
            #   "token_type": "Bearer",
            #   "scope": null,
            #   "expires_in": 1440
            #}
        #print(f"token: {self.access_token}")
        request.headers["Authorization"] = f"Bearer {self.access_token}"
        response = yield request
        #yield request
        


        # This is the json response when the token has expired
        #{
        #    "error": {
        #        "errorCode": 401,
        #        "description": "token is expired",
        #        "message": "unauthorized"
        #    }
        #}
        # VS 
        # The json response when credentials are invalid
        #{
        #    "error": "unauthorized",
        #    "description": "unauthorized request"
        #}
        # Both have a 401 status code



        # We don't call until fail, but check if the token has expired
        if response.status_code == 401:
        #if self.expiration_time < datetime.now(timezone.utc):
            logging.info("Refreshing tokens...")
            refresh_response = await self.build_refresh_request()
            self.update_tokens(refresh_response)

            request.headers["Authorization"] = f"Bearer {self.access_token}"
            yield request


    async def async_auth_flow(self, request):
        """
        Async auth flow.

        This method is called by the httpx client before sending the request.
        It will check if the access token is expired, and if it is it will
        refresh it.
        With this we shouldn't handle 401 errors here, but on the retry logic
        """
        if not self.access_token:
            response = await self.async_get_token()
            if response.is_error:
                response.raise_for_status()
            tokens = response.json()
            self.access_token = tokens["access_token"]
            self.refresh_token = tokens["refresh_token"]
            self.expiration_time = datetime.now(timezone.utc) + timedelta(seconds=tokens["expires_in"])
            
        # We don't call until fail, but check if the token has expired
        if self.expiration_time < datetime.now(timezone.utc):
            logging.info("Refreshing tokens...")
            print("Refreshing tokens...")
            refresh_response = await self.build_refresh_request()
            self.update_tokens(refresh_response)        
        
        request.headers["Authorization"] = f"Bearer {self.access_token}"
        yield request
        

    def sync_auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """
        Sync auth flow.
        """
        raise RuntimeError(
            "Cannot use a sync authentication class with httpx.AsyncClient"
        )

    def sync_get_token(self):
        """
        Get the token from the server.

        This method is used to get the token from the server, using the
        client credentials flow.
        """        
        with Client() as client:
            client_id = self.authorization["oauth_client_id"]
            secret = self.authorization["oauth_secret"]
            oauth_url = self.authorization["oauth_url"]+"/v2/oauth/token"
            token = f"{client_id}:{secret}"
            token = b64encode(str.encode(token)).decode("utf-8")
            data = {"grant_type": "client_credentials"}            
            headers = {
                "Authorization": f"Basic {token}",
                "Content-Type": "application/x-www-form-urlencoded",
            }
            response = client.post(
                oauth_url,
                data=data,
                headers=headers,
                timeout=constants.TIMEOUT,
            )
            return response

    async def async_get_token(self):
        """
        Get the token from the server.

        This method is used to get the token from the server, using the
        client credentials flow.
        """
        async with AsyncClient() as client:
            client_id = self.authorization["oauth_client_id"]
            secret = self.authorization["oauth_secret"]
            oauth_url = self.authorization["oauth_url"]+"/v2/oauth/token"
            token = f"{client_id}:{secret}"
            token = b64encode(str.encode(token)).decode("utf-8")
            data = {"grant_type": "client_credentials"}
            headers = {
                "Authorization": f"Basic {token}",
                "Content-Type": "application/x-www-form-urlencoded",
            }
            response = await client.post(
                oauth_url,
                data=data,
                headers=headers,
                timeout=constants.TIMEOUT,
            )
            return response

    async def build_refresh_request(self) -> Response:
        """
        Build the refresh request.

        This method is used to build the request to refresh the tokens.
        """
        async with AsyncClient() as client:
            client_id = self.authorization["oauth_client_id"]
            secret = self.authorization["oauth_secret"]
            oauth_url = self.authorization["oauth_url"]+"/v2/oauth/token"
            token = f"{client_id}:{secret}"
            token = b64encode(str.encode(token)).decode("utf-8")
            data = {"grant_type": "refresh_token", "refresh_token": self.refresh_token}
            headers = {
                "Authorization": f"Basic {token}",
                "Content-Type": "application/x-www-form-urlencoded",
            }
            response = await client.post(
                oauth_url,
                data=data,
                headers=headers,
                timeout=constants.TIMEOUT,
            )
            return response

    def update_tokens(self, response: Response) -> None:
        """
        Update the tokens.

        This method is used to update the tokens after a refresh request.
        """
        data = response.json()
        self.access_token = data["access_token"]
        self.refresh_token = data["refresh_token"]
        self.expiration_time = datetime.now(timezone.utc) + timedelta(seconds=data["expires_in"])
