import asyncio
import tempfile, json

from httpx import Timeout
from tenacity import retry, retry_if_result

import utils.constants as constants
from request_handler.client import Client
from utils.rate_limit import is_retryable_status_code, wait_until_next_window
from request_handler.auth import AribaAuth
from templates_methods.sync.base import BaseTemplateMethod


class BaseAsyncTemplateMethod(BaseTemplateMethod):
    """ Base class to handle asynchronous endpoints from ARIBA. """
    def __init__(self, api_settings: dict):
        super().__init__(api_settings)
        self._base_filters = api_settings.get("params", {}).get("base", {}).get("filters")
        self._wait_for_async = api_settings.get("wait_for_async", 18)       

    @retry(
        retry=retry_if_result(is_retryable_status_code),
        wait=wait_until_next_window,
        sleep=asyncio.sleep
    )
    async def async_post(self, url: str, json_data, extra_params=None):
        """ Makes a POST call with the configured settings. """
        async with Client(auth=AribaAuth(authorization=self._base_auth)) as client:
            self._logger.info(f"POST: {url}")
            if extra_params:
                self._logger.debug(f"Extra Params {extra_params}")
                params = self._base_params | extra_params
            else:
                params = self._base_params
            
            try:
                response = await client.post(
                    url=url,
                    follow_redirects=True,
                    headers=self._base_headers,
                    params=params,
                    json=json_data,
                    timeout=Timeout(30, connect=30),
                )
                response.raise_for_status()
                return response
            except constants.RETRIABLE_EXPCETIONS as e:
                return e

    @retry(
        retry=retry_if_result(is_retryable_status_code),
        wait=wait_until_next_window,
        sleep=asyncio.sleep
    )
    async def async_get(self, url: str, params={}):
        """ Makes a GET call with the configured settings. """
        async with Client(auth=AribaAuth(authorization=self._base_auth)) as client:
            self._logger.info(f"GET {url}")
            try:
                response = await client.get(
                    url=url,
                    follow_redirects=True,
                    headers=self._base_headers,
                    params=self._base_params | params,
                    timeout=Timeout(30, connect=30),
                )
                response.raise_for_status()
                return response
            except constants.RETRIABLE_EXPCETIONS as e:
                return e


    @retry(
        retry=retry_if_result(is_retryable_status_code),
        wait=wait_until_next_window,
        sleep=asyncio.sleep
    )
    async def async_stream_get_file(self, url: str, filename):
        """ Opens up a stream connection to download a file with the configured settings. """
        self._logger.info(f"GET file {url}")
        async with Client(auth=AribaAuth(authorization=self._base_auth)) as client:
            try:
                async with client.stream("GET", url, follow_redirects=True, headers=self._base_headers, params=self._base_params) as r:
                    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                        async for data in r.aiter_bytes():
                            temp_file.write(data)
                        temp_file.close()
                        return temp_file
            except constants.RETRIABLE_EXPCETIONS as e:
                return e
                


