import json
from string import Template

import httpx


from templates_methods.async_methods.base import BaseAsyncTemplateMethod
from utils.file_handling import create_folders, file_exists, full_path_file, get_data_from_file, rename, unzip_file

from asyncio import Queue
import asyncio
from utils.last_succesful_run import LastSuccessfulRun, SingletonMeta

from utils.producer_consumer import SharedCondition


class ZippedFileTemplateMethod(BaseAsyncTemplateMethod):
    """ Class that represents the template method: ZippedFile. Used to download zipped files from asynchronous endpoints from 
        Operational Reporting for Procurement API
    """

    __metaclass__ = SingletonMeta

    def __init__(self, api_settings: dict):
        super().__init__(api_settings)
        self._list_url = Template(api_settings["urls"]["list"])
        self._list_params = api_settings["params"]["list"]
        self._list_path = api_settings["paths"]["list"]
        filters = api_settings.get("params", {}).get("list", {}).get("filters").copy()
        filters_keys = api_settings.get("date_filters", None)
        # Map the user set date filter to this 
        from_key = "updatedDateFrom"
        to_key = "updatedDateTo"

        if filters_keys is not None:
            from_key = filters_keys.get("from")
            to_key = filters_keys.get("to")
            
        filters[from_key] = filters.pop("dateFrom")
        filters[to_key] = filters.pop("dateTo")
        
        self._base_filters = filters
        self._list_filters = json.dumps(filters)
        

        self._download_url = Template(api_settings["urls"]["download"])
        self._download_params = api_settings["params"]["download"]
        self._download_path = api_settings["paths"]["download"]
        self._download_filters = json.dumps(filters)
        

        self.viewTemplateName = (
            api_settings.get("body", {}).get("base", {}).get("viewTemplateName", "")
        )
        self._detail_list_key = api_settings.get("keys",{}).get("detail_list_key", None)
        self._extra_templates = None
        self.init_extras()
        


    def init_extras(self):
        if self._extras is not None:
            self._extra_templates = {}
            for extra_key, model_data in self._extras.items():
                self._extra_templates[extra_key] = model_data.get("method")(model_data)
                last_run = LastSuccessfulRun(self._name)
                self._extra_templates[extra_key].set_last_run(last_run)

    async def start_job(self, pageToken=None):
        """ Method that calls Job Submission API.

        Args:
            pageToken (string): Token to be used to call next page.

        Returns:
            job_info: Information related to the job that was initiated.
                jobId (string): uuid of the job just created
                files (list[string]): a list of fileIds
                status (string): status of the job. One of: ['pending', 'processing','completed', 'completedZeroRecords','expired']
                pageToken (string): page token to fetch next set of result.
                totalNumOfRecords (int): Total number of records.
                totalNumOfPages (int): Amount of pages of this request.
                currentPageRecordsCount (int): Amount of records for this page.
                currentPageNum (int): Number of the current page.
                totalNumOfFiles (int): Total number of files.
                filesInCurrentPage (int): Amount of files for this page.
        
        """
        url = self._base_url.substitute(self._base_path)
        data = {
            "viewTemplateName": self.viewTemplateName,
            "filters": self._base_filters,
        }
        extra_params = {"pageToken": pageToken} if pageToken else {}
        self._base_params.pop("filters", None)
        response = await self.async_post(url, data, extra_params=extra_params)
        if isinstance(response, httpx.HTTPStatusError):
            return None
        self._logger.debug(response.headers)
        json_response = response.json()
        job_info = {
            "jobId": json_response.get("jobId", ""),
            "files": json_response.get("files", ""),
            "status": json_response.get("status", ""),
            "pageToken": json_response.get("pageToken", None),
            "totalNumOfRecords": json_response.get("totalNumOfRecords", ""),
            "totalNumOfPages": json_response.get("totalNumOfPages", ""),
            "currentPageRecordsCount": json_response.get("currentPageRecordsCount",""),
            "currentPageNum": json_response.get("currentPageNum", ""),
            "totalNumOfFiles": json_response.get("totalNumOfFiles", ""),
            "filesInCurrentPage": json_response.get("filesInCurrentPage", ""),
        }
        return job_info

    async def get_job(self, jobId: str):
        """ Method that retrieves the status of a job started in 'start_job'. Also called Job Results API

        Args:
            jobId (string): Id of the job we want to get it status.

        Returns:
            job_info: Information related to the job that was initiated.
                jobId (string): uuid of the job just created
                files (list[string]): a list of fileIds
                status (string): status of the job. One of: ['pending', 'processing','completed', 'completedZeroRecords','expired']
                pageToken (string): page token to fetch next set of result.
                totalNumOfRecords (int): Total number of records.
                totalNumOfPages (int): Amount of pages of this request.
                currentPageRecordsCount (int): Amount of records for this page.
                currentPageNum (int): Number of the current page.
                totalNumOfFiles (int): Total number of files.
                filesInCurrentPage (int): Amount of files for this page. 
        """
        url = self._list_url.substitute(self._list_path | {"job_id": jobId})
        response = await self.async_get(url, params=self._list_params)
        if isinstance(response, httpx.HTTPStatusError):
            return None
        self._logger.debug(response.headers)
        data = response.json()
        job_status = {
            "jobId": data.get("jobId", ""),
            "files": data.get("files", ""),
            "status": data.get("status", ""),
            "pageToken": data.get("pageToken", ""),
            "totalNumOfRecords": data.get("totalNumOfRecords", ""),
            "totalNumOfPages": data.get("totalNumOfPages", ""),
            "currentPageRecordsCount": data.get("currentPageRecordsCount", ""),
            "currentPageNum": data.get("currentPageNum", ""),
            "totalNumOfFiles": data.get("totalNumOfFiles", ""),
            "filesInCurrentPage": data.get("filesInCurrentPage", ""),
        }
        return job_status

    async def get_file(self, jobId: str, fileId: str):
        """ Method that downloads a file from a job using the fileId and the jobId"""
        url = self._download_url.substitute(
            self._download_path | {"job_id": jobId, "file_id": fileId}
        )
        file_object = await self.async_stream_get_file(url, fileId)
        return file_object


    async def download_data(self):
        """ Main method that will download data."""
        # queue from asyncio used to share ids
        queue = Queue()
        # Defines if we should end consuming ids from the queue
        endCondition = SharedCondition(False)

        # Producer task: downloads files, reads them and insert ids into the queue
        producer = asyncio.create_task(self.download_jobs(queue, endCondition))
        # Consumer task: reads ids from the queue and download data associated with that id
        consumer = asyncio.create_task(self.consumer(queue, endCondition))

        # wait for both tasks to end.
        await asyncio.gather(producer, consumer)


    async def download_jobs(self, queue: Queue, endCondition: SharedCondition):
        self._logger.info("Starting async download")
        print("Starting zipped file download template method...")
        job_info = await self.start_job()
        if job_info is not None:
            jobId = job_info.get("jobId")
            status = job_info.get("status")
            self._logger.info(f"Job: {jobId} started...")
            print(f"Requesting job {jobId}")
            status_message = "Waiting for job complete status"
            print(status_message, end="\r")
            while status in ["pending", "processing"]:
                job_status = await self.get_job(jobId)

                if job_status is None:
                    self._logger.error("Non retriable error when getting job status.")
                    break

                status = job_status.get("status", "")            
                self._logger.info(f"Status: {status}")
                if status in ["pending", "processing"]:
                    self._logger.info("Waiting for status update...")
                    status_message = status_message + "."
                    print(status_message,end="\r")
                    # wait for X seconds to not break rate limit for the get job status endpoint
                    seconds = self._wait_for_async 
                    await asyncio.sleep(seconds)
                    self._logger.info("Retrying...")                
                    continue
                elif status == "expired":
                    self._logger.warning(f"Job {jobId} expired.")
                    break
                elif status == "completed":                
                    self._logger.info("Files ready!")
                    files = job_status.get("files", [])
                    await self.download_all_files(jobId, files, queue)
                    self._logger.info(f"{len(files)} downloaded")

                pageToken = job_status.get("pageToken", None)
                if pageToken:
                    self._logger.info("Getting next page...")
                    job_info = await self.start_job(pageToken)
                    if job_info is None:
                        self._logger.error("Non retriable error when getting next page job.")
                        break
                    jobId = job_info.get("jobId")
                    status = job_info.get("status")
        else:
            self._logger.error("Non retriable error when starting job.")

        self._logger.info("Job finished")

        endCondition.value = True

        

    async def download_all_files(self, jobId, files, queue: Queue):
        """ Method that download all files in the list of files."""
        tasks = []        
        for fileId in files:
            self._logger.info(f"Getting file: {fileId}")
            print(f"Downloading file {fileId}")
            file_object = await self.get_file(jobId, fileId)
            self._logger.info(f"File {fileId} downloaded!")
            # Creates a task to process the downloaded file concurrently, so we can start downloading the next file without waiting
            task = asyncio.create_task(self.process_downloaded_file(fileId, file_object, queue))

        #wait for all files to be processed
        for task in tasks:
            await task
            

    async def process_downloaded_file(self, fileId, file_object, queue):
        """Method to fully process recently downloaded file"""
        self._logger.info(f"Processing file: {fileId}")
        print(f"Processing file: {fileId}")

        # Store the full folder path in folder_path
        folder_path = full_path_file(self._output_dir, "zip")
        #if the folder does not exists, then create it
        if not file_exists(folder_path):
            create_folders(folder_path)

        # decrompress the downloaded zip file
        unzip_file(file_object, folder_path)

        # calculate the new full path for the decompressed file
        new_file = full_path_file(folder_path, fileId.replace(".zip", ".json"))
        #rename decompressed "records.txt" to "{fileId}.json"
        rename(full_path_file(folder_path, "records.txt"), new_file)
        
        # Try to read the json file
        try:
            file_data = get_data_from_file(new_file)
        except:
            self._logger.error(f"Error: reading file {new_file}")

        # if we defined a payload key in the settings file, use it!
        if self._payload_key:
            file_data = file_data[self._payload_key]
        
        self._logger.info(f"File {new_file} had {len(file_data)} elements")
        print(f"Adding {len(file_data)} rows to csv")

        #single model was extracted from the view template
        if self._model:
            self._logger.info(f"Parsing data to a single model")
            self.add_tables_data(file_data)
        #multiple models were extracted from the view template
        elif self._models:
            self._logger.info(f"Parsing data to multiple models")            
            self.add_tables_data_from_models(file_data)

        # check if it needs to insert ids into the queue
        if self._extras is not None:
            await self.producer(file_data, queue)
        

    async def producer(self, file_data,  queue: Queue):
        """ Method to insert into the queue the id field needed to start 'extra' processes"""
        if self._view_extractor is not None:
            id_field = self._view_extractor.get_id_field()
            if id_field is not None:
                self._logger.debug(f"There are {len(file_data)} to add")
                for item in file_data:
                    id = item.get(id_field)
                    self._logger.debug(f"Adding {id} to queue")
                    await queue.put(id)
        else:
            file_data = list(filter(lambda x: x.get("TemplateObject", None), file_data))
            for data in file_data:
                id = data.get(self._detail_list_key, None)
                if id is not None:
                    await queue.put(id)


    async def consumer(self, queue: Queue, endCondition: SharedCondition):
        """ Method that will process the elements in the queue and retrieve info from 'extras' ."""
        if self._extras is not None: 
            #This will run until we process every element in the queue, and the end condition is met.   
            while True:
                #reset temporary list of ids
                temp_list = []
                # while there are no more elements in the queue, process them
                while not queue.empty():
                    id = await queue.get()
                    self._logger.debug(f"Processing {id} from queue. Task remaining {queue.qsize()}")
                    temp_list.append(id)
                    queue.task_done()
                    # if we have 300 elements in the temporary list: -> we stop because of rate-limit of 300 calls per second in the document approval api
                    if len(temp_list) == 300:
                        break

                
                if len(temp_list) > 0:
                    # pass on the temporary list to every element in 'extras'
                    for extra_key, model_data in self._extras.items():
                        self._logger.info(f"Starting process of {extra_key} with {len(temp_list)} tasks")
                        await self._extra_templates[extra_key].download_data(temp_list.copy())

                # if we processed every element in the queue and the end condition is met, we break the infinite loop
                if endCondition.value and queue.empty():
                    break
                # if the end condition isn't met, wait for the queue to refill
                elif queue.empty():
                    await asyncio.sleep(self._wait_for_async)
                
            


