from abc import ABC, abstractmethod
import asyncio
from datetime import datetime
import os
from typing import List
from string import Template

from pydantic import BaseModel

import utils.constants as constants

from logger import logger_conf

from tenacity import retry, retry_if_result
from utils.current_run import CurrentRunConstants
from utils.file_handling import file_exists
from utils.last_succesful_run import LastSuccessfulRun

from utils.rate_limit import is_retryable_status_code, wait_until_next_window
from request_handler.client import Client
from request_handler.auth import AribaAuth


class BaseTemplateMethod(ABC):
    """Base class for template method pattern."""

    def __init__(self, api_settings: dict):
        self._name = api_settings['model_name']
        self._base_url = Template(api_settings["urls"]["base"])
        self._base_headers = api_settings["headers"]["base"]
        self._base_auth = api_settings["auth"]["base"]
        self._base_params = api_settings.get("params", {}).get("base", {})
        self._base_path = api_settings["paths"]["base"]
        self._model = api_settings.get("model", None)
        self._output_handler = api_settings.get("output_handler", None)
        self._payload_key = api_settings.get("keys", {}).get("payload_key", None)
        self._tables = api_settings["tables"]
        self._models = api_settings.get("models", None)
        self._output_dir = api_settings["output_dir"]
        self._view_extractor = api_settings.get("view_name", None)
        self._logger = logger_conf.get_logger(self._name)
        self._last_updated_attribute = api_settings.get("last_updated_attribute", None)
        self._extras = None  
        if api_settings.get("extras", None) is not None:
            print(f"Setting extras for {self._name}")
            self._extras = api_settings.get("extras")    

        self.current_run = CurrentRunConstants()
        #if not self.current_run.get_value(f"{self._name}_cleaned"):
        #    self.current_run.set_value(f"{self._name}_cleaned", True)
        #    self.clean_csv()


    def clean_csv(self):
        """
        Check if csv's already exists, if so... delete it or rename old one
        """
        if self._output_handler:
            output_handler = self._output_handler(
                tables=self._tables,
                model=self._model,
                output_dir=self._output_dir,
            )
            extension = output_handler.get_extension()
            for table_name, headers in self._tables.items():
                #full_path = os.path.join(self._output_dir, table_name)
                #file_name = f"{full_path}{extension}"
                file_name = output_handler.get_filename(table_name)
                full_file_name = f"{file_name}{extension}"
                if file_exists(full_file_name):
                    if self.current_run.get_value("remove_old"):
                        os.remove(full_file_name)
                    else:
                        timestamp = datetime.now().strftime("%Y%m%d%H%M%S%f")[:-3]
                        full_new_name = f"{file_name}_previous_to_{timestamp}{extension}"
                        os.rename(full_file_name, full_new_name)
                else:
                    #if doesn't exists create it
                    output_handler.create_file(full_file_name,headers)



    def set_last_run(self, last_run: LastSuccessfulRun):
        self.last_run:LastSuccessfulRun = last_run


    def get_last_run(self) -> LastSuccessfulRun:
        return self.last_run


    @retry(
        retry=retry_if_result(is_retryable_status_code),
        wait=wait_until_next_window,
        sleep=asyncio.sleep
    )
    async def get_async(self, url: str):
        """Get data from API."""        
        async with Client(auth=AribaAuth(authorization=self._base_auth)) as client:            
            self._logger.info(f"GET {url}")
            try:
                response = await client.get(
                    url=url,
                    headers=self._base_headers,
                    params=self._base_params,
                    timeout=constants.TIMEOUT
                )
                response.raise_for_status()
                return response
            except constants.RETRIABLE_EXPCETIONS as e:
                return e
            

    @retry(
        retry=retry_if_result(is_retryable_status_code),
        wait=wait_until_next_window,
        sleep=asyncio.sleep
    )
    async def get_custom_async(self, url: str, params, headers):
        """Get data from API."""        
        async with Client(auth=AribaAuth(authorization=self._base_auth)) as client:            
            self._logger.info(f"GET {url}")
            try:
                response = await client.get(
                    url=url,
                    headers=headers,
                    params=params,
                    timeout=constants.TIMEOUT
                )
                response.raise_for_status()
                return response
            except constants.RETRIABLE_EXPCETIONS as e:
                return e
               

    @retry(
        retry=retry_if_result(is_retryable_status_code),
        wait=wait_until_next_window,
        sleep=asyncio.sleep
    )
    async def post_async(self, url: str, json_body: dict):
        """Get data from API."""        
        async with Client(auth=AribaAuth(authorization=self._base_auth)) as client:            
            self._logger.info(f"POST {url}")
            try:
                response = await client.post(
                    url=url,
                    follow_redirects=True,
                    headers=self._base_headers,
                    params=self._base_params,
                    json=json_body,
                    timeout=constants.TIMEOUT
                )
                response.raise_for_status()
                return response
            except constants.RETRIABLE_EXPCETIONS as e:
                return e

    def add_tables_data(self, payload: List[BaseModel], detail_id: str | None = None):
        """Persists the data."""
        self._logger.debug(f"Writing info...")
        print(f"Adding {len(payload)} rows to csv")
        self._logger.debug(f"Adding {len(payload)} rows to csv")

        if self._last_updated_attribute is not None:
            for item in payload:
                if hasattr(item, self._last_updated_attribute):
                    last_updated_time = getattr(item, self._last_updated_attribute)
                    self.last_run.set_latest_update(datetime.strptime(last_updated_time, "%Y-%m-%d %H:%M:%S"))

        self._output_handler(
            tables=self._tables,
            model=self._model,
            output_dir=self._output_dir,
        ).add_rows(payload, detail_id)


    def add_tables_data_from_models(self, payload: list):
        self._logger.debug(f"Writing info...")
        """Persists the data."""
        print(f"Adding {len(payload)} rows to csv.")
        self._logger.debug(f"Adding {len(payload)} rows to csv")
        for raw_item in payload:
            last_updated_time = raw_item.get("TimeUpdated",None)
            self.last_run.set_latest_update(datetime.strptime(last_updated_time, "%Y-%m-%dT%H:%M:%SZ"))
            parsed_item = self._view_extractor.parse_item(raw_item)            

            #model_key is a shared key between tables and models
            for model_key, value in self._models.items():                
                model_output = self._output_handler(tables={model_key: self._tables[model_key]}, model = value, output_dir=self._output_dir)
                model_output.add_rows(parsed_item.get(model_key))

    @abstractmethod
    def download_data(self):
        """Download data from API."""
