"""Counted template method pattern."""
import json
from string import Template
from typing import List

import aiometer
from httpx import Response
from pydantic import parse_obj_as

from templates_methods.sync.base import BaseTemplateMethod
from request_handler.auth import AribaAuth
from request_handler.client import Client
from utils.last_succesful_run import LastSuccessfulRun, SingletonMeta


class CountBeforeTemplateMethod(BaseTemplateMethod):    
    """
    Count Before template method pattern.

    This class is used decide wether use the sync or async
    methods from Operational Reporting for Procurement API
    """     
    __metaclass__ = SingletonMeta

    def __init__(self, api_settings: dict):
        super().__init__(api_settings)   
        self._count_key = api_settings.get("keys",{}).get("count_key", None)
        filters = self._base_params["filters"].copy()
        filters["updatedDateFrom"] = filters.pop("dateFrom")
        filters["updatedDateTo"] = filters.pop("dateTo")
        self._base_params["filters"] = json.dumps(filters)
        self._threshold = api_settings.get("threshold",10000)

    async def count_api_data(self) -> int:
        """Count the data from API."""
        url = self._base_url.substitute(self._base_path)
        response = await self.get_async(url)
        if isinstance(response, Response) and response.status_code == 200:
            if self._count_key:
                return response.json()[self._count_key]
            return response.json()
        else:
            return None
    

    async def download_data(self):
        """Download data from API."""
        print("Starting count before template method...")
        number_of_records = await self.count_api_data()
        self._logger.info(f"There are {number_of_records}")
        print(f"There are {number_of_records}")
        async_model_name = ""
        sync_model_name = ""
        for template_method_key in self._extras.keys():
            if "_async" in template_method_key:
                async_model_name = template_method_key
            elif "_sync" in template_method_key:
                sync_model_name = template_method_key
        if number_of_records is not None and number_of_records > self._threshold:
            #ASYNC
            await self.start_template_method(async_model_name)
        else:
            #SYNC
            await self.start_template_method(sync_model_name)


    async def start_template_method(self, model_name):
        self._logger.info(f"Starting process of {model_name}")
        print(f"Starting process of {model_name}")
        # get model data to initiate template method
        model_data = self._extras.get(model_name)
        # instatiate template method with model data
        template_model = model_data.get("method")(model_data)
        last_run = LastSuccessfulRun(self._name)
        template_model.set_last_run(last_run)
        # run template method
        await template_model.download_data()
