"""Counted template method pattern."""
import asyncio
from string import Template
from typing import List

import aiometer
from pydantic import parse_obj_as
from tenacity import retry, retry_if_result
from httpx import Response

from templates_methods.sync.base import BaseTemplateMethod
from request_handler.auth import AribaAuth
from request_handler.client import Client
from utils.last_succesful_run import SingletonMeta
import utils.constants as constants
from utils.rate_limit import is_retryable_status_code, wait_until_next_window


class CountedTemplateMethod(BaseTemplateMethod):
    """
    Counted template method pattern.

    This class is used to download data from API
    when the API has a count endpoint.
    """

    __metaclass__ = SingletonMeta

    def __init__(self, api_settings: dict):
        super().__init__(api_settings)
        self._count_url = Template(api_settings["urls"]["count"])
        self._count_headers = api_settings["headers"]["count"]
        self._count_path = api_settings["paths"]["count"]
        self._count_params = api_settings.get("params", {}).get("count", {})
        self._count_key = api_settings.get("keys", {}).get("count_key", None)
        self._count_auth = api_settings["auth"]["count"]


    @retry(
        retry=retry_if_result(is_retryable_status_code),
        wait=wait_until_next_window,
        sleep=asyncio.sleep
    )
    async def get_counted_auth(self):
        url = self._count_url.substitute(self._count_path)
        async with Client(auth=AribaAuth(authorization=self._count_auth)) as client:
            try:
                response = await client.get(
                    url=url,
                    headers=self._count_headers,
                    params=self._count_params,
                    timeout=constants.TIMEOUT,
                )
                response.raise_for_status()
                return response
            except constants.RETRIABLE_EXPCETIONS as e:
                return e 

    async def count_api_data(self) -> int:
        """Count the data from API."""
        response = await self.get_counted_auth()
        if isinstance(response, Response) and response.status_code == 200:
            if self._count_key:
                return response.json()[self._count_key]
            return response.json()
        else:
            return None

    def generate_urls(self, count: int) -> list:
        """
        Generate urls to download data from API.

        Args:
            count (int): Number of records to download.

        Returns:
            list: List of urls.
        """
        urls = []
        base_url = self._base_url.substitute(self._base_path)
        for i in range(0, count, 1000):
            params = self._base_params.copy()
            params["$skip"] = i
            params["$top"] = 1000
            urls.append(
                base_url
                + "?"
                + "&".join(f"{key}={value}" for key, value in params.items())
            )
        return urls

    async def download_data(self):
        """Download data from API."""
        print("Starting counted template method...")
        number_of_records = await self.count_api_data()
        if number_of_records is None:
            return
        urls = self.generate_urls(number_of_records)
        async with aiometer.amap(
            self.get_async,
            urls,
            max_per_second=1,
            max_at_once=1,
        ) as responses:
            async for response in responses:
                try:
                    if response.status_code!=200:
                        print("Error: ", response.status_code)
                        continue
                    payload = response.json()
                    if self._payload_key:
                        payload = payload[self._payload_key]
                    if isinstance(payload, list):
                        payload = parse_obj_as(List[self._model], payload)
                    else:
                        payload = self._model(**payload)
                    self.add_tables_data(payload)
                except Exception:
                    print("Error: ", response.status_code)
