import asyncio
from datetime import datetime
import json
from typing import List
from string import Template

import aiometer
from httpx import Timeout
import httpx
from pydantic import parse_obj_as
from tenacity import retry, retry_if_result

from request_handler.auth import AribaAuth
from request_handler.client import Client
from templates_methods.sync.base import BaseTemplateMethod


import utils.constants as constants
from utils.last_succesful_run import SingletonMeta
from utils.rate_limit import is_retryable_status_code, wait_until_next_window


class ListDetailMethod(BaseTemplateMethod):
    """
    List detail template method pattern.

    This class is used to download data from API
    when the API has a list endpoint.
    """

    __metaclass__ = SingletonMeta

    def __init__(self, api_settings: dict):
        super().__init__(api_settings)
        self._url_list = Template(api_settings["urls"].get("list", ""))

        self._list_auth = api_settings["auth"].get("list", {})
        self._list_headers = api_settings["headers"].get("list", None)
        self._list_params = api_settings["params"].get("list", None)
        filters_keys = api_settings.get("date_filters", None)
        from_key = "updatedDateFrom"
        to_key = "updatedDateTo"
        if filters_keys is not None:
            from_key = filters_keys.get("from")
            to_key = filters_keys.get("to")
        if self._list_params:
            filters = self._list_params["filters"].copy()
            filters[from_key] = filters.pop("dateFrom")
            filters[to_key] = filters.pop("dateTo")
            self._list_params["filters"] = json.dumps(filters)
        self._list_path = api_settings["paths"].get("list", None)

        self._list_key = api_settings["keys"].get("list_key", None)
        self._detail_list_key = api_settings["keys"].get("detail_list_key", None)

        self._delta_list_url = Template(api_settings["urls"].get("delta_list", ""))
        self._delta_list_path = api_settings["paths"].get("delta_list", None)
        self._delta_list_params = api_settings["params"].get("delta_list", None)
        if self._delta_list_params:
            filters = self._delta_list_params["filters"].copy()
            filters[from_key] = filters.pop("dateFrom")
            filters[to_key] = filters.pop("dateTo")
            self._delta_list_params["filters"] = filters
        self._delta_list_headers = api_settings["headers"].get("delta_list", None)


    @retry(
        retry=retry_if_result(is_retryable_status_code),
        wait=wait_until_next_window,
        sleep=asyncio.sleep
    )
    async def get_with_list_auth(self):
        url = self._url_list.substitute(self._list_path)
        self._logger.info(f"GET {url}")
        async with Client(auth=AribaAuth(authorization=self._list_auth)) as client:
            try:
                response = await client.get(
                    url=url,
                    params=self._list_params,
                    headers=self._list_headers,
                    timeout=constants.TIMEOUT,
                )
                response.raise_for_status()
                return response
            except constants.RETRIABLE_EXPCETIONS as e:
                return e 

    async def get_delta_list(self):
        url = self._delta_list_url.substitute(self._delta_list_path)
        self._logger.info(f"GET {url}")
        startKey = "modificationDateFrom"
        params = self._delta_list_params.copy()
        startDate = params.get("filters").get("updatedDateFrom", None)
        if startDate is None:
            startKey = "createDateFrom"
            startDate = params.get("filters").get("createdDateFrom", None)
        startDate = datetime.strptime(startDate, "%Y-%m-%dT%H:%M:%SZ").timestamp()
        startDate = int(startDate * 1000)


        endKey = "modificationDateTo"
        endDate = params.get("filters").get("updatedDateTo", None)
        if endDate is None:
            endKey = "createDateTo"
            endDate = params.get("filters").get("createdDateTo", None)
        endDate = datetime.strptime(endDate, "%Y-%m-%dT%H:%M:%SZ").timestamp()
        endDate = int(endDate * 1000)

        params["$filter"] = f"({startKey} gt {startDate} and {endKey} lt {endDate})"
        params.pop("filters", None)
        response = await self.get_custom_async(url, params, self._delta_list_headers)
        if isinstance(response, httpx.HTTPStatusError):
            self._logger.error("Non retriable error when getting data.")
            return []            
        else:
            projects = response.json()
            projects = projects.get("payload",[])
            list_ids = [project.get("internalId") for project in projects]
            return list_ids
            

    async def get_list_ids(self) -> list:
        """
        Get list of ids from API.

        Returns:
            list: List of ids.
        """
        response = await self.get_with_list_auth()
        records = response.json()[self._list_key]
        if response.json().get("PageToken", False):
            while response.json().get("PageToken", False):
                await asyncio.sleep(2)
                self._list_params["PageToken"] = response.json().get("PageToken")
                response = await self.get_with_list_auth()
                records.extend(response.json()[self._list_key])
        list_ids = []
        records = list(filter(lambda x: x.get("TemplateObject", None), records))
        for record in records:
            list_ids.append(record[self._detail_list_key])
        return list_ids

    def generate_urls(self, list_ids: list) -> list:
        """
        Generate urls to download data from API.

        Args:
            list_ids (list): List of ids.

        Returns:
            list: List of urls.
        """
        urls = []
        for id in list_ids:
            path = self._base_path
            path["id"] = id
            urls.append(self._base_url.substitute(path))
        return urls

   
    async def download_data(self):
        """Download data from API."""
        print("Starting list detail template method...")
        if self._delta_list_params is not None:        
            list_ids = await self.get_delta_list()
        else:
            list_ids = await self.get_list_ids()
        urls = self.generate_urls(list_ids)
        async with aiometer.amap(
            self.get_async,
            urls,
            max_per_second=0.5,
            max_at_once=1,
        ) as results:
            index_ids = -1
            async for result in results:
                index_ids += 1
                try:
                    if isinstance(result, httpx.HTTPStatusError):
                        continue
                    payload = result.json()
                    if self._payload_key:
                        payload = payload.get(self._payload_key)
                    if isinstance(payload, list):
                        new_payload = []
                        for record in payload:
                            record["detail_id"] = list_ids[index_ids]
                            new_payload.append(record)
                        payload = parse_obj_as(List[self._model], new_payload)
                    else:
                        payload["detail_id"] = list_ids[index_ids]
                        payload = [self._model(**payload)]
                    self.add_tables_data(payload)
                except KeyError:
                    print("KeyError")