from typing import List
import aiometer
import httpx
from pydantic import parse_obj_as

from templates_methods.sync.base import BaseTemplateMethod
from utils.last_succesful_run import SingletonMeta


class ListedMethod(BaseTemplateMethod):
    """
    List template method pattern.

    This class is used to download data from API
    when the API has a list endpoint.
    """ 

    __metaclass__ = SingletonMeta  

    def generate_urls(self, list_ids: list) -> list:
        """
        Generate urls to download data from API.

        Args:
            list_ids (list): List of ids.

        Returns:
            list: List of urls.
        """
        urls = []
        for id in list_ids:
            path = self._base_path
            path["id"] = id
            urls.append(self._base_url.substitute(path))
        return urls

    async def download_data(self, ids_list:list = None):
        """Download data from API."""
        print("Starting listed template method...")

        urls = self.generate_urls(ids_list)
        self._logger.info(f"Downloading {len(urls)} project details")
        async with aiometer.amap(
            self.get_async,
            urls,
            max_per_second=5,
            max_at_once=1,
        ) as results:
            index_ids = -1
            async for result in results:
                index_ids += 1
                try:
                    detail_id = ids_list[index_ids]
                    if isinstance(result, httpx.HTTPStatusError):
                        self._logger.debug(f"Error with project '{detail_id}' ignored. Continuing with next project")
                        continue
                    payload = result.json()
                    if self._payload_key:
                        payload = payload.get(self._payload_key)
                    if isinstance(payload, list):
                        new_payload = []
                        for record in payload:
                            record["detail_id"] = detail_id
                            new_payload.append(record)
                        payload = parse_obj_as(List[self._model], new_payload)
                    else:
                        payload["detail_id"] = detail_id
                        payload = [self._model(**payload)]
                    self.add_tables_data(payload)
                except KeyError:
                    print("KeyError")
