import asyncio
import aiometer
import anyio
from httpx import Response
from pydantic import parse_obj_as
from tenacity import retry, retry_if_result
from templates_methods.sync.base import BaseTemplateMethod

from request_handler.auth import AribaAuth
from request_handler.client import Client
from utils.last_succesful_run import SingletonMeta
from utils.rate_limit import is_retryable_status_code, wait_until_next_window
import utils.constants as constants


class PaginatedByCountMethod(BaseTemplateMethod):
    """
    Paginated by count template method pattern.

    This class is used to download data from API
    when it is possible to paginate by count,
    it is also useful for endpoints that have a limit
    on the number of records it can return.
    """

    __metaclass__ = SingletonMeta

    def __init__(self, api_settings: dict):
        super().__init__(api_settings)
        # Define new auth data to be used to create a new Auth object, so it can hold it's own token different than the one used for Operational Reporting for Procurement
        self._paginated_auth = api_settings["auth"]["base"]
        #remove unused filter
        self._base_params.pop("filters", None)
        # max elements per page
        self._top = 10
        
    @retry(
        retry=retry_if_result(is_retryable_status_code),
        wait=wait_until_next_window,
        sleep=asyncio.sleep
    )   
    async def get_async_without_params(self, url: str):
        """Get data from API."""        
        
        async with Client(auth=AribaAuth(authorization=self._paginated_auth)) as client:            
            self._logger.info(f"GET {url}")
            try:
                response = await client.get(
                    url=url,
                    headers=self._base_headers,
                    timeout=constants.TIMEOUT
                )
                response.raise_for_status()
                return response
            except constants.RETRIABLE_EXPCETIONS as e:
                return e 
            except anyio.EndOfStream as e:
                return e

    def generate_urls(self, ids_list) -> list[str]:
        """
        Generate urls to download data from API.

        Args:
            reqIds (list[str]): List of records to download.

        Returns:
            list: List of urls.
        """
        urls = []
        base_url = self._base_url.substitute(self._base_path)        
        for requisition_id in ids_list:
            params = self._base_params.copy()
            params["$skip"] = 0
            params["$top"] = self._top
            params["$filter"] = params["$filter"].format(id=requisition_id)
            params["$filter"] = params["$filter"].replace(" ","%20")
            urls.append(
                base_url
                + "?"
                + "&".join(f"{key}={value}" for key, value in params.items())
            )

        return urls

    async def download_data(self, ids_list:list = None):
        """Download data from API."""
        print("Starting paginated by count template method...")       
        
        urls = self.generate_urls(ids_list)
        self._logger.info(f"Downloading {len(urls)} requisitions history")

        # Tries to download from 300 urls concurrently
        async with aiometer.amap(
            self.get_async_without_params,
            urls,
            max_per_second=300
        ) as responses:
            # each call will be available as soon as the server responses
            async for response in responses:  
                if isinstance(response, Response) and response.status_code == 200:
                    original_payload = response.json()                    
                    payload, amount = self.process_response(original_payload)   
                    self.add_tables_data(payload)

                    retrieved_data = amount
                    current_page = 0
                    skip = 0
                    exists_next_page = retrieved_data < original_payload.get("count", 1)
                    previous_url = str(response.url)

                    while exists_next_page:
                        self._logger.info(f"Getting next page for {previous_url}")
                        current_page += 1
                        new_skip = current_page * self._top
                        new_url = previous_url.replace(f"$skip={skip}",f"$skip={new_skip}")
                        skip = new_skip
                        new_response = await self.get_async_without_params(new_url)
                        payload, amount = self.process_response(new_response)
                        retrieved_data += amount

                        self.add_tables_data(payload)
                        
                        previous_url = new_url
                        exists_next_page = retrieved_data < new_response.get("count", 1)
                    
                 

    
    def process_response(self, response):
        amount = 0        
        values = response.get("value")
        response_list = []

        for value in values:
            payload = value.get(self._payload_key)
            response_list.extend(parse_obj_as(list[self._model], payload))
        amount = len(values)
        return response_list, amount
        

