import asyncio
from asyncio import Queue
import json
from typing import List
import httpx

from pydantic import parse_obj_as

from templates_methods.sync.base import BaseTemplateMethod
from utils.last_succesful_run import LastSuccessfulRun, SingletonMeta
from utils.producer_consumer import SharedCondition


class PaginatedByTokenMethod(BaseTemplateMethod):
    """
    Template method pattern for paginated API.

    This class is used to download data from API
    when the API has a pagination by token.
    """

    __metaclass__ = SingletonMeta

    def __init__(self, api_settings):
        super().__init__(api_settings)
        self._object_key = api_settings.get("keys", {}).get("object_key", None)
        filters = api_settings.get("params", {}).get("base", {}).get("filters").copy()
        filters["updatedDateFrom"] = filters.pop("dateFrom")
        filters["updatedDateTo"] = filters.pop("dateTo")
        self._base_params["filters"] = json.dumps(filters)
        self.init_extras()


    def init_extras(self):
        if self._extras is not None:
            self._extra_templates = {}
            for extra_key, model_data in self._extras.items():
                self._extra_templates[extra_key] = model_data.get("method")(model_data)
                last_run = LastSuccessfulRun(self._name)
                self._extra_templates[extra_key].set_last_run(last_run)

    def generate_url(self):
        """Generate url to download data from API."""
        return self._base_url.substitute(self._base_path)


    async def download_data(self):
        queue = Queue()
        # Defines if we should end consuming ids from the queue
        endCondition = SharedCondition(False)

        # Producer task: downloads files, reads them and insert ids into the queue
        producer = asyncio.create_task(self.produce_data(queue, endCondition))
        # Consumer task: reads ids from the queue and download data associated with that id
        consumer = asyncio.create_task(self.consume_data(queue, endCondition))

        # wait for both tasks to end.
        await asyncio.gather(producer, consumer)

    async def produce_data(self, queue, endCondition: SharedCondition):
        """
        Download data from API using PageToken to handle pagination.
        """
        print("Starting paginated by token template method...")
        url = self.generate_url()
        response = await self.get_async(url)
        if isinstance(response, httpx.HTTPStatusError):
            self._logger.error("Non retriable error when getting data.")            
        else:
            await self.processPage(response, queue)

            pageToken = response.json().get("PageToken", None)
            while pageToken is not None:
                self._base_params["PageToken"] = pageToken
                self._logger.debug(f"Calling next page with PageToken : {pageToken}")
                response = await self.get_async(url)
                if isinstance(response, httpx.HTTPStatusError):
                    self._logger.error("Non retriable error when getting next page data.")
                    break
                await self.processPage(response, queue)
                pageToken = response.json().get("PageToken", None)                
            

        # Update the end condition to let the consumer know it has to finish consume data
        endCondition.value = True
       

    async def consume_data(self, queue: Queue, endCondition: SharedCondition):
        """ Method that will process the elements in the queue and retrieve info from 'extras' ."""
        if self._extras is not None: 
            #This will run until we process every element in the queue, and the end condition is met.   
            while True:
                #reset temporary list of ids
                temp_list = []
                # while there are no more elements in the queue, process them
                while not queue.empty():
                    id = await queue.get()
                    self._logger.debug(f"Processing {id} from queue. Task remaining {queue.qsize()}")
                    temp_list.append(id)
                    queue.task_done()
                    # if we have 300 elements in the temporary list: -> we stop because of rate-limit of 300 calls per second in the document approval api
                    if len(temp_list) == 300:
                        break

                
                if len(temp_list) > 0:
                    # pass on the temporary list to every element in 'extras'
                    for extra_key, model_data in self._extras.items():
                        self._logger.info(f"Starting process of {extra_key} with {len(temp_list)} tasks")
                        await self._extra_templates[extra_key].download_data(temp_list.copy())

                # if we processed every element in the queue and the end condition is met, we break the infinite loop
                if endCondition.value and queue.empty():
                    break
                # if the end condition isn't met, wait for the queue to refill
                elif queue.empty():
                    await asyncio.sleep(10)


    async def processPage(self, response, queue: Queue):
        response = response.json()
        # if we defined a payload key in the settings file, use it!
        if self._payload_key:
            payload = response[self._payload_key]
        else:
            payload = response

        #if the payload is a list, parse it
        if isinstance(payload, list):
            # if we filter each object by a field, use 'object_key' inside 'keys'
            if self._object_key is not None:
                payload = list(filter(lambda x: x.get(self._object_key, None), payload))

            #single model was extracted from the view template
            if self._model:
                self._logger.info(f"Parsing data to a single model")
                payload = parse_obj_as(List[self._model], payload)
                self.add_tables_data(payload)
            elif self._models:
                self._logger.info(f"Parsing data to multiple models")            
                self.add_tables_data_from_models(payload)
        else:
            self._logger.info(f"Parsing data to a single model")
            payload = self._model(**payload)
            self.add_tables_data(payload)
        
        # check if it needs to insert ids into the queue
        if self._extras is not None:
            await self.enqueue_data(payload, queue)




    async def enqueue_data(self, file_data,  queue: Queue):
        """ Method to insert into the queue the id field needed to start 'extra' processes"""
        id_field = self._view_extractor.get_id_field()
        if id_field is not None:
            self._logger.debug(f"There are {len(file_data)} to add")
            for item in file_data:
                id = item.get(id_field)
                self._logger.debug(f"Adding {id} to queue")
                await queue.put(id)