import asyncio
import anyio
from tenacity import retry, retry_if_result
import pytest
import httpx

from utils.constants import HTTP_EXCEPTIONS, RETRIABLE_EXPCETIONS
from utils.rate_limit import is_retryable_status_code, wait_until_next_window
from request_handler.client import Client

pytest_plugins = ('pytest_asyncio',)

URL = "http://www.fake-url.com"


@retry(
    retry=retry_if_result(is_retryable_status_code),
    wait=wait_until_next_window,
    sleep=asyncio.sleep
)
async def async_get():
    async with Client() as client:
        try:
            response = await client.get(
                url=URL,
                timeout=httpx.Timeout(30, connect=30),
            )
            response.raise_for_status()
            return response
        except RETRIABLE_EXPCETIONS as e:
            return e
        except anyio.EndOfStream as e:
            return e
        

@pytest.mark.asyncio
async def test_429_error_day_limit(httpx_mock):
    expected_text = "data"
    custom_header = {
        "x-ratelimit-remaining-second": "1",
        "x-ratelimit-remaining-minute": "8",
        "x-ratelimit-remaining-hour":"1",
        "x-ratelimit-remaining-day": "0",
        "x-ratelimit-limit-second": "300",
        "x-ratelimit-limit-minute": "500",
        "x-ratelimit-limit-hour": "1200",
        "x-ratelimit-limit-day": "3600",
        "date": "Thu, 30 Nov 2023 23:59:59 GMT"
    }   
    httpx_mock.add_response(
        url=URL,
        status_code=429,
        headers=custom_header
    )
    httpx_mock.add_response(
        url=URL,
        status_code=200,
        text=expected_text
    )

    response = await async_get()
    assert response.text == expected_text


@pytest.mark.asyncio
async def test_429_error_hour_limit(httpx_mock):
    expected_text = "data"
    custom_header = {
        "x-ratelimit-remaining-second": "1",
        "x-ratelimit-remaining-minute": "8",
        "x-ratelimit-remaining-hour":"0",
        "x-ratelimit-remaining-day": "1",
        "x-ratelimit-limit-second": "300",
        "x-ratelimit-limit-minute": "500",
        "x-ratelimit-limit-hour": "1200",
        "x-ratelimit-limit-day": "3600",
        "date": "Thu, 30 Nov 2023 23:59:59 GMT"
    }   
    httpx_mock.add_response(
        url=URL,
        status_code=429,
        headers=custom_header
    )
    httpx_mock.add_response(
        url=URL,
        status_code=200,
        text=expected_text
    )

    response = await async_get()
    assert response.text == expected_text

@pytest.mark.asyncio
async def test_429_error_minute_limit(httpx_mock):
    expected_text = "data"
    custom_header = {
        "x-ratelimit-remaining-second": "1",
        "x-ratelimit-remaining-minute": "0",
        "x-ratelimit-remaining-hour":"1",
        "x-ratelimit-remaining-day": "1",
        "x-ratelimit-limit-second": "300",
        "x-ratelimit-limit-minute": "500",
        "x-ratelimit-limit-hour": "1200",
        "x-ratelimit-limit-day": "3600",
        "date": "Thu, 30 Nov 2023 23:59:59 GMT"
    }   
    httpx_mock.add_response(
        url=URL,
        status_code=429,
        headers=custom_header
    )
    httpx_mock.add_response(
        url=URL,
        status_code=200,
        text=expected_text
    )

    response = await async_get()
    assert response.text == expected_text


@pytest.mark.asyncio
async def test_429_error_second_limit(httpx_mock):
    expected_text = "data"
    custom_header = {
        "x-ratelimit-remaining-second": "0",
        "x-ratelimit-remaining-minute": "8",
        "x-ratelimit-remaining-hour":"1",
        "x-ratelimit-remaining-day": "1",
        "x-ratelimit-limit-second": "300",
        "x-ratelimit-limit-minute": "500",
        "x-ratelimit-limit-hour": "1200",
        "x-ratelimit-limit-day": "3600",
        "date": "Thu, 30 Nov 2023 23:59:59 GMT"
    }   
    httpx_mock.add_response(
        url=URL,
        status_code=429,
        headers=custom_header
    )
    httpx_mock.add_response(
        url=URL,
        status_code=200,
        text=expected_text
    )

    response = await async_get()
    assert response.text == expected_text



@pytest.mark.asyncio
async def test_400_error_quotaExceeded(httpx_mock):
    custom_json = {
        "code": "quotaExceeded",
        "message": "The number of matched records for specified criteria exceeds the allowed limits."
    }   
    httpx_mock.add_response(
        url=URL,
        status_code=400,
        json=custom_json
    )
    response = await async_get()
    assert isinstance(response, HTTP_EXCEPTIONS)


@pytest.mark.asyncio
async def test_503_error_service_unavailable(httpx_mock):    
    httpx_mock.add_response(
        url=URL,
        status_code=503        
    )
    response = await async_get()
    assert isinstance(response, HTTP_EXCEPTIONS)

@pytest.mark.asyncio
async def test_401_error_realm(httpx_mock):    
    custom_json = {
        "error": {
            "errorCode": 10001,
            "message": "Cannot find realm: UiPathDSAPP-T2321",
            "description": None
        }
    }  
    httpx_mock.add_response(
        url=URL,
        status_code=401,
        json= custom_json       
    )
    response = await async_get()
    assert isinstance(response, HTTP_EXCEPTIONS)



@pytest.mark.asyncio
async def test_EOS_error(httpx_mock): 
    expected_text = "data"
   
    httpx_mock.add_exception(
        anyio.EndOfStream()  
    )
    httpx_mock.add_response(
        url=URL,
        status_code=200,
        text=expected_text
    )
    response = await async_get()
    assert response.text == expected_text
    
