""" File that groups together functions related to retrying an API call. """

from datetime import datetime, timedelta
import anyio
from httpx import Response
from logger import logger_conf
import utils.constants as constants
from pprint import pprint


def is_retryable_status_code(response):
    """ Define the conditions for retrying.
        returns: True if should retry, False otherwise
    """
    logger = logger_conf.get_logger("RATE_LIMIT")
    if isinstance(response, constants.TIMEOUT_EXCEPTIONS):
        logger.debug(f"A timeout {response} error was received")
        return True
    if isinstance(response, anyio.EndOfStream):
        logger.debug(f"A EndOfStream error was received")
        return True
    if isinstance(response, constants.HTTP_EXCEPTIONS):
        logger.debug(f"A {response} error was received")
        response_ = response.response
        try:
            if response_.status_code == 400:
                logger.error("Bad Request ERROR.")
                error_info = response_.json()
                if error_info.get("code", None) == "quotaExceeded":
                    logger.warning(error_info.get("code"))
                    logger.warning(error_info.get("message", "The number of matched records for specified criteria exceeds the allowed limits."))
                    print("\nQUOTA EXCEEDED: The number of matched records for specified criteria exceeds the allowed limits.")
                    print("\n\t Please use the 'async' alternative or change the filter criteria.")
                elif error_info.get("error",{}).get("errorCode",0) == 11013:
                    logger.warning("Trying to get project history that doesn't exists.")
                    logger.warning(error_info)
                elif error_info.get("error",{}).get("errorCode",0) == 11003:
                    logger.warning("The specified full Project Id is same as Template Id")
                    logger.warning(error_info)
                else:
                    print("A BAD REQUEST error has ocurred. Probably something changed in the API that was called. Please contact your developer.")
                    pprint(error_info)
                return False
            elif response_.status_code == 401:
                error_info = response_.json()
                if error_info.get("error", None) == "unauthorized" or error_info.get("message", None) == "Invalid authentication credentials":
                    logger.warning("Unauthorized ERROR: Please check your credentials are correct.")
                    print("Unauthorized ERROR: Please check your credentials are correct.")
                if isinstance(error_info.get("error"), dict) and error_info.get("error").get("errorCode",0) == 10001:
                    print(error_info.get("error",{}).get("message","User or realm doesn't exists."))
                    logger.warning(error_info.get("error",{}).get("message","User or realm doesn't exists."))
                return False
            elif response_.status_code == 503:
                logger.error("503 ERROR: Service unavailable.")
                return False
            elif response_.status_code == 403:
                error_info = response_.json()
                if error_info.get("error",{}).get("errorCode",0) == 10109:
                    logger.warning("User isn't authorized to view/edit the project.")
                    logger.warning(error_info)
                return False
        except Exception as e:
            logger.error("An error ocurred while reading response error.")
            logger.error(e)
            return False
        return True
    
    if isinstance(response, Response):
        # Never gets called: is caught previously by HTTPStatusError
        if response.is_error:
            logger.debug(f"A {response.status_code} error was received")
            return True
        else:
            # everything worked as expected
            return False

    logger.error("Unknown error ocurred.")
    logger.error(response)
    return False


def wait_until_next_window(retry_state):
    """ Calculate how much we should wait before retrying a call."""
    logger = logger_conf.get_logger("RATE_LIMIT")
    failed_response = retry_state.outcome.result()
    if isinstance(failed_response, constants.TIMEOUT_EXCEPTIONS):
        logger.debug(f"{failed_response} TimeOutError: Waiting for 30 seconds...")
        print(f"TimeOutError: Waiting for 30 seconds.")
        return 30.0
    if isinstance(failed_response, anyio.EndOfStream):
        logger.debug(f"A EndOfStream error. Waiting 1 second.")
        return 1.0
    if isinstance(failed_response, constants.HTTP_EXCEPTIONS):
        status_code = failed_response.response.status_code
        headers = failed_response.response.headers
        if status_code == 429:
            seconds_to_wait = calculate_waiting_time(headers)
            print(f"WAITING {seconds_to_wait}")
            return float(seconds_to_wait)
        if failed_response.response.is_error:
            logger.debug(f"({status_code}) error: Waiting for 30 seconds before retrying...")        
            print(f"HTTP Exception error: Waiting for 30 seconds...")
            return 30.0
    logger.debug("Unknown error: Waiting for 30 seconds before retrying...")
    logger.debug(failed_response)
    print(f"Unknown error: Waiting for 30 seconds...")
    return 30.0


def calculate_waiting_time(headers):
    """ Calculate waiting time if there is an 429 status code error."""
    logger = logger_conf.get_logger()
    logger.info("Too Many requests. Calculating seconds to wait...")
    logger.debug(headers)    
    seconds_to_wait = 10
    # Thu, 30 Nov 2023 21:12:12 GMT
    # %a, %d %b %Y %H:%M:%S %Z
    date = datetime.strptime(headers.get("date"), "%a, %d %b %Y %H:%M:%S %Z")
    # remaining tries within this window of one second
    remaining_seconds = int(headers.get("x-ratelimit-remaining-second", constants.BIG_INTEGER))
    # remaining tries within this window of one minute
    remaining_minutes = int(headers.get("x-ratelimit-remaining-minute", constants.BIG_INTEGER))
    # remaining tries within this window of one hour
    remaining_hours = int(headers.get("x-ratelimit-remaining-hour", constants.BIG_INTEGER))
    # remaining tries within this window of one day
    remaining_days = int(headers.get("x-ratelimit-remaining-day", constants.BIG_INTEGER))

    rate_limit_second = headers.get("x-ratelimit-limit-second", constants.BIG_INTEGER)
    rate_limit_minute = headers.get("x-ratelimit-limit-minute", constants.BIG_INTEGER)
    rate_limit_hour = headers.get("x-ratelimit-limit-hour", constants.BIG_INTEGER)
    rate_limit_day = headers.get("x-ratelimit-limit-day", constants.BIG_INTEGER)
    if remaining_days == 0:
        # if i don't have retries within this day, wait until next day
        next_day = date + timedelta(days=1)
        next_day = next_day.replace(hour=0, minute=0, second=0)
        seconds_to_wait = (next_day - date).total_seconds()
        logger.info(
            f"Requests per day reached ({rate_limit_day}). Waiting {seconds_to_wait} seconds until next day {next_day.strftime('%Y/%m/%d, %H:%M:%S')} GMT"
        )
        print(f"Requests per day reached ({rate_limit_day}). Waiting {seconds_to_wait} seconds until next day {next_day.strftime('%Y/%m/%d, %H:%M:%S')} GMT")
    elif remaining_hours == 0:
        # if I don't have retries within this hour, wait until next hour
        next_hour = date + timedelta(hours=1)
        next_hour = next_hour.replace(minute=0, second=0)
        seconds_to_wait = (next_hour - date).total_seconds()
        logger.info(
            f"Requests per hour reached ({rate_limit_hour}). Waiting {seconds_to_wait} seconds until next hour {next_hour.strftime('%Y/%m/%d, %H:%M:%S')} GMT"
        )
        print(f"Requests per hour reached ({rate_limit_hour}). Waiting {seconds_to_wait} seconds until next hour {next_hour.strftime('%Y/%m/%d, %H:%M:%S')} GMT")
    elif remaining_minutes == 0:
        # if I don't have retries within this minute, wait until next minute
        seconds_to_wait = 60 - date.second
        logger.info(
            f"Requests per minute reached ({rate_limit_minute}). Waiting {seconds_to_wait} seconds until next minute"
        )
        print(f"Requests per minute reached ({rate_limit_minute}). Waiting {seconds_to_wait} seconds until next minute")
    elif remaining_seconds == 0:
        # if I don't have retries withing this second, wait 1 second
        logger.info(f"Requests per second reached ({rate_limit_second}). Waiting 1 second")
        print(f"Requests per second reached ({rate_limit_second}). Waiting 1 second")
        seconds_to_wait = 1

    #logger.info(f"Waiting for {seconds_to_wait}")
    return seconds_to_wait
